@@ -1,10 +1,9 @@
-import numpy as np
-
 from nanograd.tensor import Tensor
 import nanograd.tensor as tensor
-
 from nanograd.device import Device
 
+import numpy as np
+
 from typing import Union
 
 
@@ -68,10 +67,12 @@ class Module:
     def train(self) -> None:
         """Activates training mode for network component"""
         self.is_train = True
+        return self
 
     def eval(self) -> None:
         """Activates evaluation mode for network component"""
         self.is_train = False
+        return self
     
     def forward(self, *args):
         """Forward pass of the module"""
@@ -127,18 +128,34 @@ class Module:
     def _ensure_is_initialized(self):
         """Ensures that subclass's __init__() method ran super().__init__()"""
         if self.__dict__.get('_submodules') is None:
-            raise Exception("Module not intialized. "
-                            "Did you forget to call super().__init__()?")
+            raise Exception("Modules not initialized. Did not initialize the parent class. \
+                             Call super().__init__().")
+        if self.__dict__.get('_tensors') is None:
+            raise Exception("Tensors not initialized. Did not initialize the parent class. \
+                             Call super().__init__().")
+        if self.__dict__.get('_parameters') is None: 
+            raise Exception("Parameters not initialized. Did not initialize the parent class. \
+                             Call super().__init__().")
     
     def cpu(self):
         """Moving all tensors onto the CPU"""
-        for tensor in self._tensors.items():
-            tensor[1].cpu()
+        if len(self._tensors) > 0:
+            for tensor in self._tensors.items():
+                tensor[1].cpu()
+        if len(self._submodules) > 0:
+            for module in self._submodules.items():
+                module[1].cpu()
+        return self
         
     def gpu(self):
         """Moving all tensors onto the GPU"""
-        for tensor in self._tensors.items():
-            tensor[1].gpu()
+        if len(self._tensors) > 0:
+            for tensor in self._tensors.items():
+                tensor[1].gpu()
+        if len(self._submodules) > 0:
+            for module in self._submodules.items():
+                module[1].gpu()
+        return self
 
 
 class Sequential(Module):
@@ -193,10 +210,12 @@ class Sequential(Module):
     def gpu(self):
         for layer in self.layers:
             layer.gpu()
+        return self
     
     def cpu(self):
         for layer in self.layers:
             layer.cpu()
+        return self
 
 class Linear(Module):
     """A linear layer (aka 'fully-connected' or 'dense' layer)
@@ -265,11 +284,11 @@ class BatchNorm1d(Module):
         super().__init__()
         self.num_features = num_features
 
-        self.eps = Tensor(eps, is_parameter=False)
-        self.momentum = Tensor(momentum, is_parameter=False)
+        self.eps = eps
+        self.momentum = momentum
 
-        self.gamma = Tensor.ones((self.num_features,), requires_grad=True, is_parameter=True, name="bn_gamma")
-        self.beta = Tensor.zeros((self.num_features,), requires_grad=True, is_parameter=True, name="bn_beta")
+        self.weight = Tensor.ones((self.num_features,), requires_grad=True, is_parameter=True, name="bn_gamma")
+        self.bias = Tensor.zeros((self.num_features,), requires_grad=True, is_parameter=True, name="bn_beta")
         
         self.running_mean = Tensor.zeros((self.num_features,), requires_grad=False, is_parameter=False, name="bn_running_mean")
         self.running_var = Tensor.ones((self.num_features,), requires_grad=False, is_parameter=False, name="bn_running_var")
@@ -286,8 +305,8 @@ class BatchNorm1d(Module):
             batch_var = ((x - batch_mean.reshape(shape=[1, -1])) ** 2).mean(0)
             batch_empirical_var = ((x - batch_mean.reshape(shape=[1, -1])) ** 2).sum(0) / (x.shape[0] - 1)
 
-            self.running_mean = (1 - self.momentum) * self.running_mean + self.momentum * batch_mean
-            self.running_var = (1 - self.momentum) * self.running_var + self.momentum * batch_empirical_var
+            self.running_mean = (1. - self.momentum) * self.running_mean + self.momentum * batch_mean
+            self.running_var = (1. - self.momentum) * self.running_var + self.momentum * batch_empirical_var
 
             return self._normalize(x, batch_mean, batch_var)
         else:
@@ -305,7 +324,7 @@ class BatchNorm1d(Module):
             Tensor: Normalized tensor
         """
         x_hat = (x - mean.reshape(shape=[1, -1])) / (var + self.eps).sqrt().reshape(shape=[1, -1])
-        out = self.gamma.reshape(shape=[1, -1]) * x_hat + self.beta.reshape(shape=[1, -1])
+        out = self.weight.reshape(shape=[1, -1]) * x_hat + self.bias.reshape(shape=[1, -1])
         out.name = "bn_1d_res"
         return out
 
@@ -321,18 +340,18 @@ class BatchNorm2d(Module):
         Inherits from:
             Module (nn.module.Module)
     """
-    def __init__(self, size:int, eps:float=1e-5, momentum:float=0.1) -> None:
+    def __init__(self, num_features:int, eps:float=1e-5, momentum:float=0.1) -> None:
         super().__init__()
-        self.size = size
+        self.num_features = num_features
 
-        self.eps = Tensor(eps, is_parameter=False)
-        self.momentum = Tensor(momentum, is_parameter=False)
+        self.eps = eps
+        self.momentum = momentum
 
-        self.gamma = Tensor.ones(self.size, requires_grad=True, is_parameter=True, name="bn_gamma")
-        self.beta = Tensor.zeros(self.size, requires_grad=True, is_parameter=True, name="bn_beta")
+        self.weight = Tensor.ones(self.num_features, requires_grad=True, is_parameter=True, name="bn_gamma")
+        self.bias = Tensor.zeros(self.num_features, requires_grad=True, is_parameter=True, name="bn_beta")
 
-        self.running_mean = Tensor.zeros(self.size, requires_grad=False, is_parameter=False, name="bn_running_mean")
-        self.running_var = Tensor.ones(self.size, requires_grad=False, is_parameter=False, name="bn_running_var")
+        self.running_mean = Tensor.zeros(self.num_features, requires_grad=False, is_parameter=False, name="bn_running_mean")
+        self.running_var = Tensor.ones(self.num_features, requires_grad=False, is_parameter=False, name="bn_running_var")
 
     def forward(self, x:Tensor) -> Tensor:
         """Forward pass
@@ -345,10 +364,11 @@ class BatchNorm2d(Module):
         if self.is_train == True: 
             batch_mean = x.mean(axis=(0, 2, 3)) 
             batch_var = ((x - batch_mean.reshape(shape=[1, -1, 1, 1])) ** 2).mean(axis=(0, 2, 3)) 
-            batch_empirical_var = ((x - batch_mean.reshape(shape=[1, -1, 1, 1])) ** 2).sum(axis=(0, 2, 3)) / (x.shape[0] - 1) 
+            batch_empirical_var = (((x - batch_mean.reshape(shape=[1, -1, 1, 1])) ** 2).sum(axis=(0, 2, 3)) / 
+                                  (np.prod([x.shape[i] for i in [0, 2, 3]])-1))
 
-            self.running_mean = (1 - self.momentum) * self.running_mean + self.momentum * batch_mean
-            self.running_var = (1 - self.momentum) * self.running_var + self.momentum * batch_empirical_var
+            self.running_mean = (1. - self.momentum) * self.running_mean + self.momentum * batch_mean
+            self.running_var = (1. - self.momentum) * self.running_var + self.momentum * batch_empirical_var
 
             return self._normalize(x, batch_mean, batch_var)
         else:
@@ -365,8 +385,8 @@ class BatchNorm2d(Module):
             Returns:
                 Tensor: Normalized tensor
         """
-        x_hat = (x - mean.reshape(shape=[1, -1, 1, 1])) / (var + self.eps).sqrt().reshape(shape=[1, -1, 1, 1])
-        out = self.gamma.reshape(shape=[1, -1, 1, 1]) * x_hat + self.beta.reshape(shape=[1, -1, 1, 1])
+        x_hat = (x - mean.reshape(shape=[1, -1, 1, 1])) * self.weight.reshape(shape=[1, -1, 1, 1])
+        out = x_hat / ((var + self.eps).reshape(shape=[1, -1, 1, 1]).sqrt()) + self.bias.reshape(shape=[1, -1, 1, 1])
         out.name = 'bn_2d_res'
         return out
 
@@ -387,8 +407,7 @@ class Conv1d(Module):
             Module (nn.module.Module)
     """
     def __init__(self, in_channel:int, out_channel:int, kernel_size:int, stride:int=1, 
-                 padding:Union[tuple, int]=(0, 0), weight_initialization:str="kaiming_normal",
-                 with_bias:bool=True) -> None:
+                 padding:Union[tuple, int]=(0, 0), weight_initialization:str="kaiming_normal") -> None:
         super().__init__()
         assert isinstance(padding, (tuple, int)), 'Wrong padding type. Must be integer or tuple of integers'
         if isinstance(padding, (int)): padding = (padding, padding)
@@ -398,15 +417,10 @@ class Conv1d(Module):
         self.stride, self.padding = stride, padding
         self.kernel_size = kernel_size
         self.weight_initialization = weight_initialization
-        self.with_bias = with_bias
 
         shape = (self.out_channel, self.in_channel, self.kernel_size)
-        self.weight = init_weights(shape, self.weight_initialization, 
-                                   requires_grad=True, is_parameter=True, name="conv_weight_1d")
-
-        if self.with_bias:
-            self.bias = Tensor.zeros(out_channel, requires_grad=True, 
-                                     is_parameter=True, name="conv_bias_1d")
+        self.weight = init_weights(shape, self.weight_initialization, requires_grad=True, is_parameter=True, name="conv_weight_1d")
+        self.bias = Tensor.zeros(out_channel, requires_grad=True, is_parameter=True, name="conv_bias_1d")
 
     def forward(self, x:Tensor) -> Tensor:
         """Forward pass
@@ -417,10 +431,7 @@ class Conv1d(Module):
                 Tensor: (batch_size, out_channel, output_length)
         """
         x_padded = x.pad1d(self.padding)
-
-        if self.with_bias:
-            return x_padded.conv1d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1])
-        return x_padded.conv1d(self.weight, self.stride)
+        return x_padded.conv1d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1])
 
 
 class Conv2d(Module):
@@ -439,9 +450,8 @@ class Conv2d(Module):
         Inherits from:
             Module (nn.module.Module)
     """
-    def __init__(self, in_channel:int, out_channel:int, kernel_size:int, stride:int=1, 
-                 padding:tuple=(0, 0, 0, 0), weight_initialization:str="kaiming_normal",
-                 with_bias:bool=True) -> None:
+    def __init__(self, in_channel:int, out_channel:int, kernel_size:tuple, stride:tuple=1, 
+                 padding:tuple=(0, 0, 0, 0), weight_initialization:str="kaiming_normal") -> None:
         super().__init__()
         assert isinstance(padding, (tuple, int)), 'Wrong padding type. Must be integer or tuple of integers'
         if isinstance(padding, (int)): padding = (padding, padding, padding, padding)
@@ -450,16 +460,10 @@ class Conv2d(Module):
         self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
         self.stride, self.padding = stride, padding
         self.weight_initialization = weight_initialization
-        self.with_bias = with_bias
 
         shape = (self.out_channel, self.in_channel, *self.kernel_size)
-
-        self.weight = init_weights(shape, self.weight_initialization, requires_grad=True, 
-                                   is_parameter=True, name="conv_weight_2d")
-        
-        if self.with_bias:
-            self.bias = Tensor.zeros(self.out_channel, requires_grad=True, 
-                                    is_parameter=True, name="conv_bias_2d")
+        self.weight = init_weights(shape, self.weight_initialization, requires_grad=True, is_parameter=True, name="conv_weight_2d")
+        self.bias = Tensor.zeros(self.out_channel, requires_grad=True, is_parameter=True, name="conv_bias_2d")
     
     def forward(self, x:Tensor) -> Tensor:
         """
@@ -469,25 +473,78 @@ class Conv2d(Module):
                 Tensor: (batch_size, out_channel, output_dim1, output_dim2)
         """
         x_padded = x.pad2d(self.padding)
-        if self.with_bias:
-            return x_padded.conv2d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1, 1])
-        return x_padded.conv2d(self.weight, self.stride)
+        return x_padded.conv2d(self.weight, self.stride) + self.bias.reshape(shape=[1, -1, 1, 1])
+
+
+class MaxPool1d(Module):
+    """Performs a max pooling operation after a 1d convolution
+        
+        Args:
+            pool_size (int): Kernel size
+            stride (int): Stride
+        
+        Inherits from:
+            Module (nn.module.Module)
+    """
+    def __init__(self, pool_size:int) -> None:
+        super().__init__()
+        self.pool_size = pool_size
+    
+    def forward(self, x:Tensor) -> Tensor:
+        """Performs a max pooling operation after a 1d convolution
+
+        Args:
+            x (Tensor): (batch_size, channel, in_width, in_height)
+
+        Returns:
+            Tensor: (batch_size, channel, out_width, out_height)
+        """
+        out = x.max_pool1d(self.pool_size)
+        out.name = 'mpool1d_res'
+        return out
+
+
+class AvgPool1d(Module):
+    """Performs an average pooling operation after a 1d convolution
+        
+        Args:
+            pool_size (int): Kernel size
+            stride (int): Stride
+        
+        Inherits from:
+            Module (nn.module.Module)
+    """
+    def __init__(self, pool_size:int) -> None:
+        super().__init__()
+        self.pool_size = pool_size
+    
+    def forward(self, x:Tensor) -> Tensor:
+        """Performs an average pooling operation after a 1d convolution
+
+        Args:
+            x (Tensor): (batch_size, channel, in_width, in_height)
+
+        Returns:
+            Tensor: (batch_size, channel, out_width, out_height)
+        """
+        out = x.avg_pool1d(self.pool_size)
+        out.name = 'avgpool1d_res'
+        return out
 
 
 class MaxPool2d(Module):
     """Performs a max pooling operation after a 2d convolution
     
         Args:
-            kernel_size (tuple): Kernel size
+            pool_size (tuple): Kernel size
             stride (int): Stride
         
         Inherits from:
             Module (nn.module.Module)
     """
-    def __init__(self, kernel_size:tuple, stride:int=2) -> None:
+    def __init__(self, pool_size:tuple) -> None:
         super().__init__()
-        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
-        self.stride = stride
+        self.pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
     
     def forward(self, x:Tensor) -> Tensor:
         """
@@ -497,7 +554,7 @@ class MaxPool2d(Module):
             Returns:
                 Tensor: (batch_size, channel, out_width, out_height)
         """
-        out = x.max_pool2d(self.kernel_size)
+        out = x.max_pool2d(self.pool_size)
         out.name = 'mpool2d_res'
         return out
 
@@ -506,16 +563,15 @@ class AvgPool2d(Module):
     """Performs an average pooling operation after a 2d convolution
 
         Args:
-            kernel_size (tuple): Kernel size
+            pool_size (tuple): Kernel size
             stride (int): Stride
         
         Inherits from:
             Module (nn.module.Module)
     """
-    def __init__(self, kernel_size:tuple, stride:int=2) -> None:
+    def __init__(self, pool_size:tuple) -> None:
         super().__init__()
-        self.kernel_size = kernel_size if isinstance(kernel_size, tuple) else (kernel_size, kernel_size)
-        self.stride = stride
+        self.pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
     
     def forward(self, x:Tensor) -> Tensor:
         """
@@ -525,7 +581,7 @@ class AvgPool2d(Module):
             Returns:
                 Tensor: (batch_size, channel, out_width, out_channel)
         """
-        out = x.avg_pool2d(self.kernel_size)
+        out = x.avg_pool2d(self.pool_size)
         out.name = 'avgpool2d_res'
         return out
 
@@ -563,8 +619,7 @@ class Flatten(Module):
             Returns:
                 out (Tensor): (batch_size, dim_2 * dim_3 * ...) batch_size, then all other dims flattened
         """
-        dim1, dim2 = x.shape[0], np.prod(x.shape[1:])
-        out = x.reshape((dim1, dim2))
+        out = x.flatten()
         out.name = 'flatten_res'
         return out
 
@@ -604,33 +659,32 @@ class Dropout(Module):
         return x * Tensor(1 - self.p, name='dropout_prob', is_parameter=True)
 
 
-class CrossEntropyLoss(Module):
-    """CrossEntropyLoss layer
+class NLLLoss(Module):
+    """Negative Log-likelihood loss. 
+    
+       Useful for multi-class classifiation problem.
 
-        During implementation of a model, you can either choose a CrossEntropyLoss layer
-        or call the CrossEntropy function from nn.functional.
-        
-        >>> criterion = CrossEntropyLoss()
-        >>> criterion(outputs, labels)
-        3.241
-
-        Inherits from:
-            Module (nn.module.Module)
+       The input given through a forward call is expected to contain the log-probabilities
+       of each class. To obtain these log-probabilities, simply call the log-softmax method
+       on the output vector of your model.
     """
     def __init__(self) -> None:
         pass
 
-    def forward(self, predicted:Tensor, target:Tensor) -> Tensor:
+    def forward(self, log_probs:Tensor, target:Tensor) -> Tensor:
         """Forward pass
 
-           Args:
-                predicted (Tensor): (batch_size, num_classes)
-                target (Tensor): (batch_size,)
-           Returns:
-                Tensor: loss, stored as a float in a tensor 
+        Args:
+            log_probs (Tensor): (batch_size, num_classes)
+            target (Tensor): (batch_size, 1)
+
+        Returns:
+            Tensor: loss, stored as a float in a Tensor object 
         """
-        return tensor.cross_entropy(predicted, target)
-    
+        batch_size, num_classes = log_probs.shape
+        labels = target.one_hot(num_classes)
+        return -(log_probs * labels).mean()
+        
 
 class MSELoss(Module):
     """The MSELoss function.
@@ -715,7 +769,7 @@ class LeakyReLU(Module):
             Returns:
                 Tensor: (batch_size, num_features)
         """
-        return x.relu() + self.alpha * (-x).relu()
+        return x.relu() - self.alpha * (-x).relu()
     
 
 class Sigmoid(Module):
@@ -761,37 +815,6 @@ class Tanh(Module):
         """
         return x.tanh()
 
-
-class ELU(Module):
-    r"""
-        ELU Activation Layer
-        
-        Applies an ELU (Exponential Linear Unit) activation function 
-        to the input
-
-        Alpha is learnt by gradient descent
-
-        Inherits from:
-            Module (nn.module.Module)
-    """
-    def __init__(self, alpha:float=1.0) -> None:
-        super().__init__()
-
-        self.alpha = Tensor(alpha, requires_grad=True, is_parameter=True, name="elu_alpha")
-    
-    def forward(self, x:Tensor) -> Tensor:
-        """
-            Args:
-                x (Tensor): (batch_size, num_features)
-            Returns:
-                Tensor: (batch_size, num_features)
-        """
-        pos_part = x.relu()
-        neg_part = (-x).relu()
-
-        return pos_part + self.alpha * (neg_part.exp() - 1)
-
-
 class Swish(Module):
     r"""
         Swish Activation Layer
