@@ -412,7 +412,8 @@ def matmul_forward(ctx, queue, a, b):
     return matmul_op(ctx, queue, a, b)
 
 def log_forward(ctx, queue, a):
-    return unary_op(ctx, queue, 'log(a)', a)
+    out = unary_op(ctx, queue, 'log(a)', a)
+    return out
 
 def exp_forward(ctx, queue, a):
     return unary_op(ctx, queue, 'exp(a)', a)
@@ -442,17 +443,17 @@ def reshape_forward(ctx, queue, a, shape):
     shape = tuple(-np.prod(a.shape) // np.prod(shape) if s == -1 else s for s in shape)
     return GPUBuffer(ctx, shape, hostbuf=a.data)
 
-def max_forward(ctx, queue, a, axis, keepdims):
+def max_forward(ctx, queue, a, axis):
     return reduce_op(ctx, queue, 'out = max(a, out)', 'out',
-                     a, axis=axis, keepdims=keepdims, start='-INFINITY')
+                     a, axis=axis, keepdims=False, start='-INFINITY')
 
-def min_forward(ctx, queue, a, axis, keepdims):
+def min_forward(ctx, queue, a, axis):
     return reduce_op(ctx, queue, 'out = min(a, out)', 'out',
-                     a, axis=axis, keepdims=keepdims, start='+INFINITY')
+                     a, axis=axis, keepdims=False, start='+INFINITY')
 
-def sum_forward(ctx, queue, a, axis, keepdims):
+def sum_forward(ctx, queue, a, axis):
     return reduce_op(ctx, queue, 'out += a', 'out', 
-                     a, axis=axis, keepdims=keepdims)
+                     a, axis=axis, keepdims=False)
 
 def conv1d_forward(ctx, queue, a, weight, stride):
     batch_size, in_channel, length = a.shape
@@ -554,7 +555,8 @@ def min_backward(ctx, queue, grad_output, inp, out, axis):
 def sum_backward(ctx, queue, grad_output, a, axis):
     axis = [axis] if type(axis) == int else axis
     shape = [1 if axis is None or i in axis else a.shape[i] for i in range(len(a.shape))]
-    return GPUBuffer(ctx, shape, hostbuf=grad_output)
+    output = GPUBuffer(ctx, shape, hostbuf=grad_output)
+    return element_wise_binary_op(ctx, queue, 'a+b', output, GPUBuffer(ctx, a.shape))
 
 def conv1d_backward(ctx, queue, grad_output, x, weight, stride):
     batch_size, _, out_length = grad_output.shape
@@ -697,4 +699,4 @@ def conv2d_backward(ctx, queue, grad_output, x, weight, stride):
     prgm_grad_x.conv_backward_x(queue, [batch_size, in_channel], None, *grad_x_args)
     prgm_grad_weight.conv_backward_weight(queue, [num_filters * in_channel, kernel_height, kernel_width], None, *grad_w_args)
 
-    return grad_x, grad_weight
+    return grad_x, grad_weight
\ No newline at end of file
