@@ -6,17 +6,14 @@ from nanograd.viz.comp_graph import ForwardGraphVisualizer, BackwardGraphVisuali
 
 from nanograd.device import Device
 from nanograd import autograd_engine
-
 from nanograd.autograd_engine import Function
+from nanograd.nn.ops_gpu import GPUBuffer
 
 import nanograd.nn.ops_cpu as ops_cpu
 import nanograd.nn.ops_gpu as ops_gpu
-from nanograd.nn.conv_ops import (get_conv1d_output_size, get_conv2d_output_size,
+from nanograd.nn.conv_ops import (get_conv1d_output_size, get_conv2d_output_size, 
                                   get_im2col_indices, col2im)
 
-from nanograd.nn.ops_gpu import GPUBuffer
-
-
 try:
     import pyopencl as cl
     PYOPENCL_AVAILABLE = True
@@ -116,17 +113,17 @@ class Tensor:
     @classmethod
     def zeros(cls, *shape, **kwargs):
         """Creates a Tensor filled with zeros"""
-        return cls(np.zeros(*shape), **kwargs)
+        return cls(np.zeros(*shape, dtype=np.float32), **kwargs)
     
     @classmethod
     def ones(cls, *shape, **kwargs):
         """Creates a Tensor filled with ones"""
-        return cls(np.ones(*shape), **kwargs)
+        return cls(np.ones(*shape, dtype=np.float32), **kwargs)
     
     @classmethod
     def arange(cls, *interval, **kwargs):
         """Creates a Tensor filled with values in range"""
-        return cls(np.arange(*interval), **kwargs)
+        return cls(np.arange(*interval).astype(np.float32), **kwargs)
     
     @classmethod
     def randn(cls, *shape, **kwargs):
@@ -134,7 +131,7 @@ class Tensor:
             Creates a Tensor filled with values drawn from the
             standard Gaussian distribution
         """
-        return cls(np.random.randn(*shape), **kwargs)
+        return cls(np.random.randn(*shape).astype(np.float32), **kwargs)
     
     @classmethod
     def normal(cls, loc, scale, *shape, **kwargs):
@@ -142,7 +139,7 @@ class Tensor:
             Creates a Tensor filled with values drawn from the
             a custom Gaussian distribution
         """
-        return cls(np.random.normal(loc, scale, *shape), **kwargs)
+        return cls(np.random.normal(loc, scale, *shape).astype(np.float32), **kwargs)
 
     @classmethod
     def randint(cls, low, high, *shape, **kwargs):
@@ -150,7 +147,7 @@ class Tensor:
             Creates a Tensor filled with integer values drawn in the
             range between low and high
         """
-        return cls(np.random.randint(low, high, *shape), **kwargs)
+        return cls(np.random.randint(low, high, *shape).astype(np.float32), **kwargs)
     
     @classmethod
     def eye(cls, dim, **kwargs):
@@ -158,7 +155,7 @@ class Tensor:
             Creates a 2-dimensional Tensor (matrix) equal to the
             identity matrix.
         """
-        return cls(np.eye(dim), **kwargs)
+        return cls(np.eye(dim).astype(np.float32), **kwargs)
     
     # ****************************************
     # *********** CPU/GPU support ************
@@ -186,7 +183,7 @@ class Tensor:
             return cpu_data
         
         if not isinstance(data, np.ndarray):
-            data = np.array([data]).astype(np.float32)
+            data = np.array(data, dtype=np.float32)
 
         if device == Device.GPU:
             if cl_ctx is None:
@@ -207,12 +204,14 @@ class Tensor:
 
     def cpu(self):
         self.to(Device.CPU)
+        return self
     
     def gpu(self):
         if not PYOPENCL_AVAILABLE:
             raise Exception("OpenCL is not installed in this environment. Please consider running \
                              pip install pyopencl to benefit from GPU-accelerated computations.")
         self.to(Device.GPU)
+        return self
         
 
     # ****************************************
@@ -242,7 +241,8 @@ class Tensor:
     def __str__(self):
         return f"<NanoTensor({str(self.data)}, " + \
                f"grad_fn={self.grad_fn.__class__.__name__  if self.grad_fn else None}), " + \
-               f"name={self.name}>"
+               f"name={self.name}," + \
+               f"device={self.device}>"
     
     def __repr__(self):
         return self.__str__()
@@ -262,12 +262,12 @@ class Tensor:
 
             if start is None:
                 start = 0
-            elif type(start) != int:
+            elif not np.issubdtype(type(start), int):
                 raise TypeError(f"Indices must be integer. Got {type(start)}")
             
             if stop is None:
                 stop = self.shape[i]
-            elif type(stop) != int:
+            elif not np.issubdtype(type(stop), int):
                 raise TypeError(f"Indices must be integer. Got {type(stop)}")
             elif stop < 0:
                 stop = self.shape[i] + stop
@@ -312,7 +312,7 @@ class Tensor:
         return self * other
     
     def __truediv__(self, other):
-        return self * (other ** (-1.0))
+        return self * (other ** -1.0)
     
     def __matmul__(self, other):
         return MatMul.apply(self, other, cl_ctx=cl_ctx, cl_queue=cl_queue)
@@ -321,8 +321,8 @@ class Tensor:
     # ******** Advanced operations ***********
     # ****************************************
     
-    def sum(self, axis=None, keepdims:bool=False):
-        return Sum.apply(self, axis, keepdims, cl_ctx=cl_ctx, cl_queue=cl_queue)
+    def sum(self, axis=None):
+        return Sum.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
     
     def mean(self, axis=None, keepdims:bool=False):
         out = self.sum(axis=axis)
@@ -335,11 +335,11 @@ class Tensor:
     def T(self):
         return Transpose.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
 
-    def max(self, axis=None, keepdims=False):
-        return Max.apply(self, axis, keepdims, cl_ctx=cl_ctx, cl_queue=cl_queue)
+    def max(self, axis=None):
+        return Max.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
     
-    def min(self, axis=None, keepdims=False):
-        return Min.apply(self, axis, keepdims, cl_ctx=cl_ctx, cl_queue=cl_queue)
+    def min(self, axis=None):
+        return Min.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
     
     def log(self):
         return Log.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
@@ -363,6 +363,11 @@ class Tensor:
     def squeeze(self, axis):
         return Squeeze.apply(self, axis, cl_ctx=cl_ctx, cl_queue=cl_queue)
 
+    def flatten(self):
+        dim1, dim2 = self.shape[0], np.prod(self.shape[1:])
+        out = self.reshape((dim1, dim2))
+        return out
+
     # ****************************************
     # ******** Activation functions **********
     # ****************************************
@@ -376,48 +381,86 @@ class Tensor:
     def tanh(self):
         return Tanh.apply(self, cl_ctx=cl_ctx, cl_queue=cl_queue)
     
+    def log_softmax(self):
+        batch_size, num_classes = self.shape
+        a = self.max(axis=1).reshape((batch_size, 1)) # Log-exp trick
+        out = self - a - (self - a).exp().sum(axis=1).unsqueeze(0).log().reshape((batch_size, 1))
+        return out
+
     # ****************************************
     # ********* Conv/Pool operations *********
     # ****************************************
+
+    def _pool1d(self, field_length:int):
+        """1-dimensional pooling operation
+
+        Args:
+            field_length (int): length of the pooling kernel
+        
+        Returns:
+            x_reshaped (Tensor): pooled Tensor
+        """
+        x_unpadded = self[:, :, :self.shape[2] - self.shape[2] % field_length]
+
+        shape = (x_unpadded.shape[0], x_unpadded.shape[1], 
+                 x_unpadded.shape[2] // field_length, field_length)
+
+        x_reshaped = x_unpadded.reshape(shape=shape)
+        return x_reshaped 
     
-    def _pool2d(self, field_height:int, field_width:int):
-        r"""
+    def max_pool1d(self, kernel_size:int=2):
+        """MaxPooling1d operation
+
+        Args:
+            kernel_size (int, optional): Kernel length for pooling operation. Defaults to 2.
+        """
+        return self._pool1d(kernel_size).max(axis=3)
+    
+    def avg_pool1d(self, kernel_size:int=2):
+        """AvgPooling1d operation
+
+        Args:
+            kernel_size (int, optional): Kernel length for pooling operation. Defaults to 2.
+        """
+        return self._pool1d(kernel_size).mean(axis=3)
+        
+    def _pool2d(self, pool_size:tuple):
+        """
             2-dimensional pooling operation
 
             Args:
-                field_height (int): height of the pooling kernel
-                field_width (int): width of the pooling kernel
+                pool_size (tuple): height of the pooling kernel
             
             Returns:
                 x_reshaped (Tensor): pooled Tensor
         """
         x_unpadded = self[:, :, 
-            :self.shape[2] - self.shape[2] % field_height, 
-            :self.shape[3] - self.shape[3] % field_width]
+            :self.shape[2] - self.shape[2] % pool_size[0], 
+            :self.shape[3] - self.shape[3] % pool_size[1]]
         
         shape = (x_unpadded.shape[0], x_unpadded.shape[1], 
-                 x_unpadded.shape[2] // field_height, field_height, 
-                 x_unpadded.shape[3] // field_width, field_width)
+                 x_unpadded.shape[2] // pool_size[0], pool_size[0], 
+                 x_unpadded.shape[3] // pool_size[1], pool_size[1])
 
         x_reshaped = x_unpadded.reshape(shape=shape)
 
         return x_reshaped
     
-    def max_pool2d(self, kernel_size:tuple=(2, 2)):
+    def max_pool2d(self, pool_size:tuple=(2, 2)):
         """MaxPooling2d operation
         
             Args:
                 kernel_size (tuple): Kernel length for pooling operation
         """
-        return self._pool2d(*kernel_size).max(axis=5).max(axis=3)
+        return self._pool2d(pool_size).max(axis=(3, 5))
     
-    def avg_pool2d(self, kernel_size:tuple=(2, 2)):
+    def avg_pool2d(self, pool_size:tuple=(2, 2)):
         """AvgPooling2d operation
         
             Args:
                 kernel_size (tuple): Kernel length for pooling operation
         """
-        return self._pool2d(*kernel_size).mean(axis=(3, 5))
+        return self._pool2d(pool_size).mean(axis=(3, 5))
 
     def pad1d(self, pad:tuple):
         """Padding for one-dimensional signal
@@ -428,7 +471,7 @@ class Tensor:
         Returns:
             Tensor: Padded signal
         """
-        return self[:, :, -pad[0]:self.shape[2]+pad[1]]
+        return self[:, :, -pad[0]:int(self.shape[2])+pad[1]]
 
     def pad2d(self, pad:tuple):
         """Padding for two-dimensional images
@@ -440,7 +483,7 @@ class Tensor:
         Returns:
             Tensor: Padded signal
         """
-        return self[:, :, -pad[2]:self.shape[2]+pad[3], -pad[0]:self.shape[3]+pad[1]]
+        return self[:, :, -pad[2]:int(self.shape[2])+pad[3], -pad[0]:int(self.shape[3])+pad[1]]
 
     def conv1d(self, weight, stride:int):
         """1d convolution
@@ -486,30 +529,9 @@ class Tensor:
 
 
 # ****************************************
-# ************* Functional ***************
+# ************** Functional **************
 # ****************************************
 
-def cross_entropy(predicted:Tensor, target:Tensor) -> Tensor:
-    """Calculates Cross Entropy Loss between logits and true labels.
-       Used in the CrossEntropy module
-
-    Args:
-        predicted (Tensor): Logits
-        target (Tensor): Target classes
-
-    Returns:
-        Tensor: Loss in a Tensor of shape ()
-    """
-    batch_size, num_classes = predicted.shape
-    labels = target.one_hot(num_classes)
-
-    a = predicted.max(axis=1).reshape((batch_size, 1))
-    log_softmax = predicted - a - (predicted - a).exp().sum(axis=1).log().reshape((batch_size, 1))
-    nll_loss = - (log_softmax * labels).sum() / batch_size
-
-    return nll_loss
-
-
 class OneHot(Function):
     @staticmethod
     def forward(ctx, a, num_classes):
@@ -694,7 +716,7 @@ class Reshape(Function):
 
 class Max(Function):
     @staticmethod
-    def forward(ctx, a, axis=None, keepdims=False):
+    def forward(ctx, a, axis=None):
         axis = [axis] if type(axis) == int else axis
 
         ctx.save_for_backward(a)
@@ -703,10 +725,9 @@ class Max(Function):
         is_leaf = not requires_grad
 
         if a.device == Device.CPU:
-            out_data = ops_cpu.max_forward(a.data, axis, keepdims)
+            out_data = ops_cpu.max_forward(a.data, axis)
         else:
-            out_data = ops_gpu.max_forward(ctx.cl_ctx, ctx.cl_queue, 
-                                           a.data, axis, keepdims)
+            out_data = ops_gpu.max_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)
         
         ctx.axis, ctx.out = axis, out_data
 
@@ -733,7 +754,7 @@ class Max(Function):
 
 class Min(Function):
     @staticmethod
-    def forward(ctx, a, axis=None, keepdims=False):
+    def forward(ctx, a, axis=None):
         axis = [axis] if type(axis) == int else axis
 
         ctx.save_for_backward(a)
@@ -742,9 +763,9 @@ class Min(Function):
         is_leaf = not requires_grad
 
         if a.device == Device.CPU:
-            out_data = ops_cpu.min_forward(a.data, axis, keepdims)
+            out_data = ops_cpu.min_forward(a.data, axis)
         else:
-            out_data = ops_gpu.min_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis, keepdims)
+            out_data = ops_gpu.min_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)
         
         ctx.axis, ctx.out = axis, out_data
 
@@ -774,14 +795,14 @@ class Log(Function):
     def forward(ctx, a):
         ctx.save_for_backward(a)
         requires_grad = a.requires_grad
-
+    
         if a.device == Device.CPU:
             out_data = ops_cpu.log_forward(a.data)
         else:
             out_data = ops_gpu.log_forward(ctx.cl_ctx, ctx.cl_queue, a.data)
 
         out = Tensor(out_data, requires_grad=requires_grad, 
-                            is_leaf=not requires_grad, device=a.device)
+                     is_leaf=not requires_grad, device=a.device)
         out.children = [a]   
         out.op = 'log'                
         return out
@@ -864,7 +885,7 @@ class Add(Function):
 
 class Sum(Function):
     @staticmethod
-    def forward(ctx, a, axis, keepdims):
+    def forward(ctx, a, axis):
         ctx.axis = axis
         ctx.save_for_backward(a)
 
@@ -872,12 +893,12 @@ class Sum(Function):
         is_leaf = not requires_grad
 
         if a.device == Device.CPU:
-            out_data = ops_cpu.sum_forward(a.data, axis, keepdims)
+            out_data = ops_cpu.sum_forward(a.data, axis)
         else:
-            out_data = ops_gpu.sum_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis, keepdims)
+            out_data = ops_gpu.sum_forward(ctx.cl_ctx, ctx.cl_queue, a.data, axis)
 
         out = Tensor(out_data, requires_grad=requires_grad, 
-                            is_leaf=is_leaf, device=a.device)
+                     is_leaf=is_leaf, device=a.device)
         out.children = [a]
         out.op = 'sum'
         return out
@@ -1012,7 +1033,7 @@ class Mul(Function):
             out_data = ops_gpu.mul_forward(ctx.cl_ctx, ctx.cl_queue, a.data, b.data)
 
         out = Tensor(out_data, requires_grad=requires_grad, 
-                            is_leaf=not requires_grad, device=a.device)
+                     is_leaf=not requires_grad, device=a.device)
         out.children = [a, b]
         out.op = 'mul'
         return out
@@ -1129,8 +1150,8 @@ class Conv1d(Function):
         is_leaf = not x.requires_grad
 
         if x.device == Device.CPU:
-            out, x_cols = ops_cpu.conv1d_forward(x.data, weight.data, stride)
-            ctx.x_cols = x_cols
+            out, x_reshaped = ops_cpu.conv1d_forward(x.data, weight.data, stride)
+            ctx.x_reshaped = x_reshaped
         else:
             out = ops_gpu.conv1d_forward(ctx.cl_ctx, ctx.cl_queue, x.data, weight.data, stride)
         
@@ -1148,8 +1169,8 @@ class Conv1d(Function):
         stride = ctx.stride
 
         if grad_output.device == Device.CPU:
-            x_cols = ctx.x_cols
-            grad_x, grad_weight = ops_cpu.conv1d_backward(grad_output.data, x.data, x_cols, weight.data, stride)
+            x_reshaped = ctx.x_reshaped
+            grad_x, grad_weight = ops_cpu.conv1d_backward(grad_output.data, x, x_reshaped, weight.data, stride)
         else:
             grad_x, grad_weight = ops_gpu.conv1d_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
                                                           x.data, weight.data, stride)
@@ -1164,8 +1185,8 @@ class Conv2d(Function):
         is_leaf = not requires_grad
 
         if x.device == Device.CPU:
-            out, x_cols = ops_cpu.conv2d_forward(x.data, weight, stride)
-            ctx.x_cols = x_cols
+            out, x_reshaped = ops_cpu.conv2d_forward(x.data, weight.data, stride)
+            ctx.x_reshaped = x_reshaped
         else:
             out = ops_gpu.conv2d_forward(ctx.cl_ctx, ctx.cl_queue, x.data, weight.data, stride)
 
@@ -1184,11 +1205,11 @@ class Conv2d(Function):
         stride = ctx.stride
 
         if grad_output.device == Device.CPU:
-            x_cols = ctx.x_cols
-            grad_x, grad_weight = ops_cpu.conv2d_backward(grad_output.data, x, weight, x_cols, stride)
+            x_reshaped = ctx.x_reshaped
+            grad_x, grad_weight = ops_cpu.conv2d_backward(grad_output.data, x, x_reshaped, weight.data, stride)
         else:
             grad_x, grad_weight = ops_gpu.conv2d_backward(ctx.cl_ctx, ctx.cl_queue, grad_output.data, 
-                                                                     x.data, weight.data, stride)
+                                                          x.data, weight.data, stride)
 
         grad_x = Tensor(grad_x, device=grad_output.device)
         grad_weight = Tensor(grad_weight, device=grad_output.device)
